package ru.dbotthepony.kstarbound.defs.world

import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.fromJson
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.world.Universe
import ru.dbotthepony.kstarbound.world.UniversePos
import ru.dbotthepony.kstarbound.world.WorldGeometry
import kotlin.properties.Delegates

class WorldTemplate(val geometry: WorldGeometry) {
	var seed: Long = 0L
		private set
	var worldName: String = ""
		private set
	var worldParameters: VisitableWorldParameters? = null
		private set
	var worldLayout: WorldLayout? = null
		private set
	var skyParameters: SkyParameters = SkyParameters()
		private set
	var celestialParameters: CelestialParameters? = null
		private set

	constructor(worldParameters: VisitableWorldParameters, skyParameters: SkyParameters, seed: Long) : this(WorldGeometry(worldParameters.worldSize, true, false)) {
		this.seed = seed
		this.skyParameters = skyParameters
		this.worldLayout = worldParameters.createLayout(seed)
	}

	fun determineName() {
		if (celestialParameters != null) {
			worldName = celestialParameters!!.name
		} else if (worldParameters is FloatingDungeonWorldParameters) {

		} else {
			worldName = ""
		}
	}

	@JsonFactory
	data class SerializedForm(
		val celestialParameters: CelestialParameters? = null,
		val worldParameters: VisitableWorldParameters? = null,
		val skyParameters: SkyParameters = SkyParameters(),
		val seed: Long = 0L,
		val size: Either<WorldGeometry, Vector2i>,
		val regionData: JsonElement = JsonNull.INSTANCE,
		//val customTerrainRegions:
	)

	fun toJson(isLegacy: Boolean): JsonObject {
		val data = Starbound.gson.toJsonTree(SerializedForm(
			celestialParameters, worldParameters, skyParameters, seed,
			if (isLegacy) Either.right(geometry.size) else Either.left(geometry),
		)) as JsonObject

		data["regionData"] = worldLayout?.toJson(isLegacy) ?: JsonNull.INSTANCE
		return data
	}

	companion object {
		suspend fun create(coordinate: UniversePos, universe: Universe): WorldTemplate {
			val params = universe.parameters(coordinate) ?: throw IllegalArgumentException("$universe has nothing at $coordinate!")
			val visitable = params.visitableParameters ?: throw IllegalArgumentException("$coordinate of $universe is not visitable")

			val template = WorldTemplate(WorldGeometry(visitable.worldSize, true, false))

			template.seed = params.seed
			template.worldLayout = visitable.createLayout(params.seed)
			template.skyParameters = SkyParameters.create(coordinate, universe)
			template.celestialParameters = params
			template.worldParameters = visitable

			template.determineName()

			return template
		}

		fun fromJson(data: JsonObject): WorldTemplate {
			val load = Starbound.gson.fromJson(data, SerializedForm::class.java)
			val template = WorldTemplate(load.size.map({ it }, { WorldGeometry(it, true, false) }))

			template.celestialParameters = load.celestialParameters
			template.worldParameters = load.worldParameters
			template.skyParameters = load.skyParameters
			template.seed = load.seed
			template.worldLayout = load.regionData.let { if (it is JsonObject) WorldLayout().fromJson(it) else null }

			template.determineName()

			return template
		}
	}
}
