package ru.dbotthepony.kstarbound.defs.world.terrain

import com.google.gson.Gson
import com.google.gson.JsonObject
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kommons.gson.consumeNull
import ru.dbotthepony.kommons.gson.value
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound

fun createNamedTerrainSelector(name: String, parameters: TerrainSelectorParameters): AbstractTerrainSelector<*> {
	return Registries.terrainSelectors.getOrThrow(name).value.create(parameters)
}

enum class TerrainSelectorType(val jsonName: String) {
	CONSTANT("constant"),
	DISPLACEMENT("displacement"),
	;

	companion object : TypeAdapter<AbstractTerrainSelector<*>>(), TypeAdapterFactory {
		override fun <T : Any> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
			if (AbstractTerrainSelector::class.java.isAssignableFrom(type.rawType)) {
				return this as TypeAdapter<T>
			}

			return null
		}

		override fun write(out: JsonWriter, value: AbstractTerrainSelector<*>?) {
			if (value == null)
				out.nullValue()
			else
				out.value(value.toJson())
		}

		private val objects by lazy { Starbound.gson.getAdapter(JsonObject::class.java) }

		override fun read(`in`: JsonReader): AbstractTerrainSelector<*>? {
			if (`in`.consumeNull())
				return null

			return create(objects.read(`in`))
		}

		fun createFactory(json: JsonObject): TerrainSelectorFactory<*, *> {
			val name = json["name"]?.asString ?: ""
			val type = json["type"]?.asString?.lowercase() ?: throw JsonSyntaxException("Missing 'type' element of terrain json")

			when (type) {
				CONSTANT.jsonName -> {
					return TerrainSelectorFactory(name, Starbound.gson.fromJson(json, ConstantTerrainSelector.Data::class.java), ::ConstantTerrainSelector)
				}

				DISPLACEMENT.jsonName -> {
					return TerrainSelectorFactory(name, Starbound.gson.fromJson(json, DisplacementTerrainSelector.Data::class.java), ::DisplacementTerrainSelector)
				}

				else -> throw IllegalArgumentException("Unknown terrain selector type $type")
			}
		}

		fun create(json: JsonObject): AbstractTerrainSelector<*> {
			return createFactory(json).create(Starbound.gson.fromJson(json["parameters"] ?: throw JsonSyntaxException("Missing 'parameters' element of terrain json"), TerrainSelectorParameters::class.java))
		}

		fun create(json: JsonObject, parameters: TerrainSelectorParameters): AbstractTerrainSelector<*> {
			return createFactory(json).create(parameters)
		}
	}
}
