package ru.dbotthepony.kstarbound.util

import it.unimi.dsi.fastutil.objects.Object2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.objects.Object2ObjectMap
import it.unimi.dsi.fastutil.objects.Object2ObjectMaps

class Directives private constructor(private val directivesInternal: Object2ObjectAVLTreeMap<String, String>) {
	constructor() : this(Object2ObjectAVLTreeMap())
	constructor(directives: String) : this() {
		if (directives.isNotBlank()) {
			if ('?' !in directives) {
				if ('=' !in directives) {
					throw IllegalArgumentException("Missing render directive delimiter in '$directives'")
				}

				// assume it is just "name=value"
				val key = directives.substringBefore('=')
				val value = directives.substringAfter('=')
				directivesInternal[key] = value
			} else {
				// gets interesting
				for (pair in directives.split('?').filter { it.isNotBlank() }) {
					if ('=' !in pair) {
						throw IllegalArgumentException("Missing render directive delimiter in '$pair' (full string: '$directives')")
					}

					val key = pair.substringBefore('=')
					val value = pair.substringAfter('=')
					directivesInternal[key] = value
				}
			}
		}
	}

	val directives: Object2ObjectMap<String, String> = Object2ObjectMaps.unmodifiable(directivesInternal)

	override fun toString(): String {
		if (directivesInternal.isEmpty())
			return "Directives[empty]"
		else
			return "Directives[?${directivesInternal.entries.joinToString("?") { "${it.key}=${it.value}" }}]"
	}

	override fun equals(other: Any?): Boolean {
		return this === other || other is Directives && directivesInternal == other.directivesInternal
	}

	override fun hashCode(): Int {
		return directivesInternal.hashCode()
	}

	fun add(directive: String, value: String): Directives {
		if (directivesInternal[directive] == value)
			return this

		val copy = directivesInternal.clone()
		copy[directive] = value
		return Directives(copy)
	}

	fun add(directives: String): Directives {
		if ('?' !in directives) {
			if ('=' !in directives) {
				throw IllegalArgumentException("Missing render directive delimiter in $directives")
			}

			// assume it is just "name=value"
			val key = directives.substringBefore('=')
			val value = directives.substringAfter('=')
			return add(key, value)
		} else {
			// gets interesting
			val split = directives.split('?').filter { it.isNotBlank() }

			if (split.isEmpty())
				return this

			val copy = directivesInternal.clone()

			for (pair in split) {
				if ('=' !in pair) {
					throw IllegalArgumentException("Missing render directive delimiter in '$pair' (full string: $directives)")
				}

				val key = pair.substringBefore('=')
				val value = pair.substringAfter('=')
				copy[key] = value
			}

			return Directives(copy)
		}
	}
}
